// DrivingMotor.gs
// 
// significant parts of the vehicle light code have been derived from
// that in lights.gs by David (Marinus) Hamann
//
// SetLocoNumber written by David Taylor

include "locomotive.gs"

class DrivingTrailer isclass Locomotive {

  thread void DirectionMonitor();
  thread void VehicleMonitor();
  void SetLocoNumber(int nClass, int nLower, int nUpper);
  int IsEndCar();
  void SetLights(int cartype, bool direction);

  public define int CAR_DERAILED = -1;
  public define int CAR_CENTER = 0;
  public define int CAR_FRONT = 1;
  public define int CAR_BACK = 2;
  public define int CAR_SINGLE = 3;

  Asset coronawhite, coronared, meAsset;
  int status;
  bool runDM = true;

  //The Init function is called when the object is created
  public void Init() {
    // call the parent

    inherited();
//    Interface.Log("Init start (inherited called)");

    SetLocoNumber(369,001,005);

    meAsset = GetAsset();

    // set up corona KUID stuff

    KUID coronawhiteKUID = GetAsset().LookupKUIDTable("corona-white");
    KUID coronaredKUID = GetAsset().LookupKUIDTable("corona-red");
    coronawhite = World.FindAsset(coronawhiteKUID);
    coronared = World.FindAsset(coronaredKUID);

    // set up message sniffing

    Train mytrain = me.GetMyTrain();
    Sniff(mytrain, "Train", "", true);

    // set off our monitoring threads

    VehicleMonitor(); //Start the Couple Monitor

//    Interface.Log("Init end");
  }

  // Monitoring the couple events. If a couple event happens, we check if we are now / are no longer
  // the last car in the consist
  thread void VehicleMonitor() {
    Message msg;
      
    Interface.Log("Couple Monitor start");
    while(status != CAR_DERAILED) {
      status = IsEndCar();
      if(status and runDM) {
        runDM = false;
        DirectionMonitor();
      }

//      Interface.Log("Couple Monitor Waiting");
      wait() {
        on "Vehicle", "Coupled": {
//          Interface.Log("VehicleMonitor: Vehicle Coupled");
        }
        on "Vehicle", "Decoupled": {
//          Interface.Log("VehicleMonitor: Vehicle Decoupled");
        }
        on "Vehicle", "BadCouple": {
//          Interface.Log("VehicleMonitor: Vehicle BadCouple");
        }
        on "Vehicle", "Derailed", msg: {
          if(msg.src == me) {
//            Interface.Log("VehicleMonitor: I am derailed");
            status = CAR_DERAILED;
          }
        }
        on "Train", "StartedMoving": {
//          Interface.Log("VehicleMonitor: Train started moving");
        }
      }
    }
  }

  // The Direction Monitor checks every x seconds, in which direction our Vehicle is facing
  thread void DirectionMonitor() {
//    Interface.Log("Direction Monitor start");
    bool direction;

    while(status > CAR_CENTER) // Exit when we are no longer an end car
    {
//      Interface.Log("Direction Monitor loop");
      status = IsEndCar();
      direction = GetFacingRelativeToTrain();
      SetLights(status, direction);
      Sleep(1);
    }

    direction = GetFacingRelativeToTrain();
    SetLights(status, direction);

    runDM = true;
 
//    Interface.Log("Direction Monitor end");
  }

  int IsEndCar() {
    Train consist;
    Vehicle[] cars;
    int rval = CAR_CENTER;

    consist = me.GetMyTrain();
    cars = consist.GetVehicles();
    if(me == cars[0]) {
      rval = rval + CAR_FRONT;
    }
    if(me == cars[cars.size() - 1]) {
      rval = rval + CAR_BACK;
    }
    return rval;
  }


  void SetLights(int cartype, bool direction) {

    Interface.Log("SetLights called with cartype: \"" + cartype + "\" and direction: \"" + direction + "\"");

    switch(cartype) {

      case CAR_DERAILED:
      case CAR_CENTER:
//        Interface.Log("Switching off all lights - car is center car");
        SetFXCoronaTexture("0",null);
        SetFXCoronaTexture("1",null);
        SetFXCoronaTexture("2",null);
        SetFXCoronaTexture("3",null);
        break;

      case CAR_FRONT:
        if(direction) {
//          Interface.Log("Switching on no.2 end marker lights - car is front car headed in opposite direction to train");
          SetFXCoronaTexture("0",null);
          SetFXCoronaTexture("1",null);
          SetFXCoronaTexture("2",null);
          SetFXCoronaTexture("3",null);
        } else {
//          Interface.Log("Switching on no.1 end marker lights - car is front car headed in same direction as train");
          SetFXCoronaTexture("0",coronawhite);
          SetFXCoronaTexture("1",coronawhite);
          SetFXCoronaTexture("2",null);
          SetFXCoronaTexture("3",null);
        }
        break;

      case CAR_BACK:
        if(direction) {
//          Interface.Log("Switching on no.1 end tail lights - car is end car headed in opposite direction to train");
          SetFXCoronaTexture("0",null);
          SetFXCoronaTexture("1",null);
          SetFXCoronaTexture("2",coronared);
          SetFXCoronaTexture("3",coronared);
        } else {
//          Interface.Log("Switching on no.2 end tail lights - car is end car headed in same direction as train");
          SetFXCoronaTexture("0",null);
          SetFXCoronaTexture("1",null);
          SetFXCoronaTexture("2",null);
          SetFXCoronaTexture("3",null);
        }
        break;

      case CAR_SINGLE:
        if(direction) {
//          Interface.Log("Switching on no.2 end marker lights - car is single car headed in opposite direction to train");
          // If I had a no.2 end, I'd be setting the no.2 end marker lights too...
          SetFXCoronaTexture("0",null);
          SetFXCoronaTexture("1",null);
          SetFXCoronaTexture("2",coronared);
          SetFXCoronaTexture("3",coronared);
        } else {
//          Interface.Log("Switching on no.1 end marker lights - car is single car headed in same direction to train");
          // If I had a no.2 end, I'd be setting the no.2 end tail lights too...
          SetFXCoronaTexture("0",coronawhite);
          SetFXCoronaTexture("1",coronawhite);
          SetFXCoronaTexture("2",null);
          SetFXCoronaTexture("3",null);
        }
        break;

      default:
        // this can't happen
        Interface.Log("Broken Code alert, someone called SetLights with a cartype of \"" + cartype + "\" and direction of \"" + direction + "\"");
        break;

    }
  }


  void SetLocoNumber(int nClass, int nLower, int nUpper) {

    // Strip off the class number
    string strCurrentNumber = me.GetRunningNumber();
    if (strCurrentNumber != "") {

      // Assume that whatever number has been entered is what the
      // user wants, even if it is not prototypically correct.
      return;
    }

    // Build a list of all the possible numbers for this class.  We'll remove
    // those numbers already allocated later on in the function.
    int nCount;
    int [] arLocoNumbers  = new int[nUpper - nLower + 1];
    for (nCount = 0; nCount < (nUpper - nLower + 1); nCount++) {
      arLocoNumbers[nCount] = nLower + nCount;
    }

    // Now get all the locos in the session, check for a class and number match,
    // and flag them as being used in the loco list by setting the number to -1.
    Vehicle[] arVehicles = World.GetVehicleList();
    for (nCount = 0; nCount < arVehicles.size(); nCount++) {

      Vehicle vCurrent = arVehicles[nCount];
      if (vCurrent.GetEngineType() != Vehicle.ENGINE_NONE) {

        // It's a loco - is it the same class ?
        string strWorldNumber = vCurrent.GetRunningNumber();
        if (strWorldNumber != "" and 
            strWorldNumber.size() > (cast<string>nClass).size() and
            strWorldNumber[0, (cast<string>nClass).size()] == cast<string> nClass) {

          // Strip off the class number by grabbing the rightmost digits
          Str.Right(strWorldNumber, strWorldNumber.size() - (cast<string> nClass).size());
          int nWorldNumber = Str.ToInt(strWorldNumber);
          if (nWorldNumber >= nLower and nWorldNumber <= nUpper) {

            // Got a match - mark it as "in use".  Note that this relies on the 1 to 1
            // relationship between the array index and the actual running number.
            // Hence we can't just delete the matching array element as future
            // deletions will get messed up (as the index would be wrong).
            arLocoNumbers[nWorldNumber - nLower] = -1;
          }
        }
      }
    }
    
    // Make a single contiguous array of available numbers (ie copy the LocoNumbers
    // array, skipping the -1 entries)
    int nAvailableCount = 0;
    int [] arAvailableNumbers;
    for (nCount = 0; nCount < arLocoNumbers.size(); nCount++) {

      if (arLocoNumbers[nCount] != -1) {
        // Number is available
        if (arAvailableNumbers == null) {
          arAvailableNumbers = new int[1];
        }
        arAvailableNumbers[nAvailableCount] = arLocoNumbers[nCount];
        nAvailableCount++;
      }
    }

    // At this stage, arAvailableNumbers should be (total number in class - those already allocated)
    string strFinal;
    if (arAvailableNumbers != null) {
      // Now do the random number.  Note that we don't use rand to explicitly generate
      // a random road number.  Instead, we use rand to generate a random index into
      // the available numbers array, and use whatever number is sitting at that index.
      strFinal = Str.ToInt(arAvailableNumbers[Math.Rand(0, arAvailableNumbers.size())]);
    } else {
      // All numbers allocated, so just generate a duplicate within the range
      strFinal = Math.Rand(nLower, nUpper + 1);
    }
    
    if (strFinal.size() == 1) {
      // single digit - need to add two 0's (xx003, xxx005 etc)
      strFinal = "00" + strFinal;
    }
    if (strFinal.size() == 2) {
      // two digits - need to add one 0 (xx023, xxx055 etc)
      strFinal = "0" + strFinal;
    }
    strFinal = cast<string> nClass + strFinal;
    
    me.SetRunningNumber(strFinal);
  }

};
