include "locomotive.gs"

class Sparker isclass Locomotive {

  thread void make_spark(string attachment, float duration);
  thread void spark_manager();
  Asset spark;

  //The Init function is called when the object is created
  public void Init() {

    inherited(); // Calls the Init function of the parent Class (Locomotive)
//    Interface.Log("Sparker: Init called");

    spark = me.GetAsset().FindAsset("spark");

    if (me.GetMyTrain()) {
//      Interface.Log("Sparker: We have a train");

      // train is defined, should be OK to run

      make_spark("0", 0);

      make_spark("1", 0);

      make_spark("2", 0);

      make_spark("3", 0);

      make_spark("4", 0);

      make_spark("5", 0);

      make_spark("6", 0);

      make_spark("7", 0);

      spark_manager();

    } else {

//      Interface.Log("Sparker: We do NOT have a train");

    }

//    Interface.Log("Sparker: Init finished");
  }

  thread void spark_manager() {

    Interface.Log("Sparker: spark_manager called");

    int pantos_up;
    float power;
    float speed;
    int weather = 1;
    int sleep_done = 0;
    float time_to_sleep;

    for(;;) { // loop forever

      pantos_up = me.GetMyTrain().GetPantographState();
//      Interface.Log("Sparker: pantograph state: " + pantos_up + ".");

      if (pantos_up) {

        // get weather
        switch(World.GetWeatherType()) {
          case World.WEATHER_TYPE_CLEAR : weather = 1; break;
          case World.WEATHER_TYPE_CLOUDY : weather = 1; break;
          case World.WEATHER_TYPE_DRIZZLE : weather = 2; break;
          case World.WEATHER_TYPE_RAIN : weather = 4; break;
          case World.WEATHER_TYPE_STORMY : weather = 6; break;
          case World.WEATHER_TYPE_LIGHT_SNOW : weather = 2; break;
          case World.WEATHER_TYPE_MEDIUM_SNOW : weather = 5; break;
          case World.WEATHER_TYPE_HEAVY_SNOW : weather = 8; break;
          default : weather = 1;
        }

        speed = Math.Fabs(me.GetVelocity());

        // get physics model
        if(me.GetMyTrain().GetPhysicsModel() == Train.PHYSICS_DCC) {

          // speed is used extensively here to try to reproduce an approximation of the behaviour
          // in CAB mode, as we can't get the ammeter in DCC mode.

          // This constant will need tuning. It is the speed at which to start tailing off the
          // sparking in m/s. I've used 10 here, which is about 35mph, on a loco capable of 90.
          // This is to reproduce the approximate behaviour of reading the ammeter in cab mode,
          // so that trains running at full speed don't spark away really madly when they wouldn't
          // in CAB mode. Trains running at lower speeds will either have the DCC throttle set lower,
          // which will reduce the sparking, or will be working hard at a grade, so will spark
          // enthusiastically, which is what we want.

          if(speed > 10) {
            speed = Math.Sqrt(10 / speed);
          } else if(speed > 1) {
            speed = 1.0;
          }

          power = Math.Fabs(me.GetMyTrain().GetDCCThrottle());

        } else {

          // speed is only used to quieten the sparks down when not moving very much, the CAB mode
          // calculation is based very heavily on the ammeter reading

          if(speed > 1) {
            speed = 1.0;
          }

          // this constant will need tuning. It is the square root of the expected sustained ammeter
          // reading during periods of relatively heavy load, for example working up a grade on full
          // throttle at (say) 20 to 30 mph. The square root of the ammeter peak on acceleration can
          // be higher, but not by much - you wouldn't want the 'power' variable to stay much above '1'
          // on a sustained basis.
          power = Math.Sqrt(GetEngineParam("current-drawn")) / 20;

        }

//        Interface.Print("Power: " + power + " Speed: " + speed);

        // avoid div-zero infinite
        if(power <= 0.01) {
          power = 0.01;
        }
        if(speed <= 0.01) {
          speed = 0.01;
        }
      }

      time_to_sleep = 10;
      if(pantos_up) {

        // tweak this Rand() call if you prefer more, or less sparks overall. The values should be
        // as far apart as possible to make it unpredictable. Regular flashing is something we'd like
        // to avoid. Under full power, at low to mid speeds, in good weather, this will be about the rate
        // of sparks, as the divisor will tend to 1 in these conditions. Bad weather makes more sparks,
        // working the engine less makes less sparks.

        time_to_sleep = Math.Rand(0.25,1.75) / (power * speed * weather);
      }
//      Interface.Print("Sleep time: " + time_to_sleep + " Done: " + sleep_done);

      if(time_to_sleep - sleep_done < 1) {

        switch (pantos_up) {
          case 1 :

            // attachment points for front panto

            make_spark(Math.Rand(0,8), Math.Rand(0.01,0.1));

            break;

          case 2 :

            // attachment points for rear panto, which we don't have, so is commented out

            // make_spark(Math.Rand(8,16), Math.Rand(0.01,0.1));

            break;

          case 3 :

            // attachment points for both pantos, but we don't have a back panto, so I've
            // commented this out, and just copied the front panto line in here instead

            make_spark(Math.Rand(0,8), Math.Rand(0.01,0.1));
            // make_spark(Math.Rand(0,16), Math.Rand(0.01,0.1));

            break;

          default : 

            // no pantos up, do nothing

            break;            
        }

        Sleep(time_to_sleep - sleep_done);
        sleep_done = 0;

      } else {

        Sleep(1);
        sleep_done++;

      }
    }
  }


  thread void make_spark(string attachment, float duration) {

//    Interface.Log("make_spark called for attachment " + attachment + " with duration " + duration + " seconds.");

    SetFXCoronaTexture(attachment, spark);
    Sleep(duration);
    SetFXCoronaTexture(attachment, null);

//    Interface.Log("make_spark finished for attachment " + attachment + " with duration " + duration + " seconds.");
  }
};


